/*********************************************************************************
 * Copyright (c) 2006-2008 Forschungszentrum Juelich GmbH 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * (1) Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the disclaimer at the end. Redistributions in
 * binary form must reproduce the above copyright notice, this list of
 * conditions and the following disclaimer in the documentation and/or other
 * materials provided with the distribution.
 * 
 * (2) Neither the name of Forschungszentrum Juelich GmbH nor the names of its 
 * contributors may be used to endorse or promote products derived from this 
 * software without specific prior written permission.
 * 
 * DISCLAIMER
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 ********************************************************************************/

package eu.unicore.gateway.base;

import java.security.cert.CertificateException;
import java.security.cert.CertificateExpiredException;
import java.security.cert.X509Certificate;

import javax.net.ssl.X509TrustManager;

import org.apache.log4j.Logger;

import eu.unicore.gateway.properties.GatewayProperties;
import eu.unicore.gateway.util.LogUtil;
import eu.unicore.security.CertificateUtils;

/**
 * <p>
 * AuthSSLX509TrustManager can be used to extend the default
 * {@link X509TrustManager} with additional trust checks. One example would be a
 * CRL check
 * </p>
 */
public class AuthSSLX509TrustManager implements X509TrustManager {

	
	private X509TrustManager defaultTrustManager = null;
	private GatewayProperties props;

	private static final Logger LOG = LogUtil.getLogger(LogUtil.SECURITY,AuthSSLX509TrustManager.class);

	/**
	 * Constructor for AuthSSLX509TrustManager.
	 */
	public AuthSSLX509TrustManager(final X509TrustManager defaultTrustManager, GatewayProperties props) {
		super();
		if (defaultTrustManager == null) {
			throw new IllegalArgumentException("Trust manager may not be null");
		}
		this.defaultTrustManager = defaultTrustManager;
		this.props = props;
	}
	

	public void checkClientTrusted(X509Certificate[] certificates, String s)
			throws CertificateException {
		logCerts("Client", certificates);
		
		try{
			//check CRLs
			CertificateUtils.verifyCertificate(certificates, true, false);
			defaultTrustManager.checkClientTrusted(certificates, s);
		}
		catch(CertificateExpiredException iae){
			String msg="Certificate "+CertificateUtils.safePrintSubject(certificates)+" is no longer valid.";
			CertificateExpiredException cee=new CertificateExpiredException(msg);
			cee.initCause(iae);
			throw cee;
		}
		catch(CertificateException ce){

			if(props.isProxyEnabled()){
				// Attempt to validate Proxy X.509 Certificate
				LOG.debug("Authenticating X.509 Proxy Certificate");
				IProxyCertificateValidator validator = ProxyValidatorFactory.getProxyValidatorInstance(props.getProxyValidator());
				try{
					validator.validate(certificates, getAcceptedIssuers());
				}catch(Exception e){
					throw new CertificateException("The supplied certificate "+CertificateUtils.safePrintSubject(certificates)+
							" could not be authenticated (both normal and proxy mode were checked.", e);
				}
			}
			else {
				throw new CertificateException("The supplied certificate "+CertificateUtils.safePrintSubject(certificates)+
						" could not be authenticated!", ce);
			}
		}
	}

	/**
	 * @see X509TrustManager#isServerTrusted(X509Certificate[])
	 */
	public void checkServerTrusted(X509Certificate[] certificates, String s)
			throws CertificateException {
		logCerts("Server", certificates);
		defaultTrustManager.checkServerTrusted(certificates, s);
		//check CRLs
		CertificateUtils.verifyCertificate(certificates, true, false);
	}

	/**
	 * @see X509TrustManager#getAcceptedIssuers()
	 */
	public X509Certificate[] getAcceptedIssuers() {
		return this.defaultTrustManager.getAcceptedIssuers();
	}
	
	private void logCerts(String type, X509Certificate[] certificates)
	{
		if (LOG.isDebugEnabled() && certificates != null) {
			StringBuilder sb = new StringBuilder();
			for (int c = 0; c < certificates.length; c++) {
				X509Certificate cert = certificates[c];
				sb.append(type);
				sb.append(" certificate " + (c + 1));
				sb.append(":\n  Subject DN: " + cert.getSubjectDN());
				sb.append("\n  Signature Algorithm: " + cert.getSigAlgName());
				sb.append("\n  Valid from: " + cert.getNotBefore());
				sb.append("\n  Valid until: " + cert.getNotAfter());
				sb.append("\n  Issuer: " + cert.getIssuerDN());
			}
			LOG.debug(sb.toString());
		}
	}
}
