/*********************************************************************************
 * Copyright (c) 2006 Forschungszentrum Juelich GmbH 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * (1) Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the disclaimer at the end. Redistributions in
 * binary form must reproduce the above copyright notice, this list of
 * conditions and the following disclaimer in the documentation and/or other
 * materials provided with the distribution.
 * 
 * (2) Neither the name of Forschungszentrum Juelich GmbH nor the names of its 
 * contributors may be used to endorse or promote products derived from this 
 * software without specific prior written permission.
 * 
 * DISCLAIMER
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 ********************************************************************************/


package eu.unicore.gateway.base;

import java.net.URI;
import java.net.URL;

import org.apache.log4j.Logger;
import org.mortbay.jetty.AbstractConnector;
import org.mortbay.jetty.Connector;
import org.mortbay.jetty.Handler;
import org.mortbay.jetty.Server;
import org.mortbay.jetty.SessionIdManager;
import org.mortbay.jetty.ajp.Ajp13SocketConnector;
import org.mortbay.jetty.servlet.Context;
import org.mortbay.jetty.servlet.DefaultServlet;
import org.mortbay.jetty.servlet.FilterHolder;
import org.mortbay.jetty.servlet.HashSessionIdManager;
import org.mortbay.jetty.servlet.ServletHolder;
import org.mortbay.thread.QueuedThreadPool;

import eu.unicore.gateway.Gateway;
import eu.unicore.gateway.properties.GatewayProperties;
import eu.unicore.gateway.properties.JettyProperties;
import eu.unicore.gateway.properties.SecurityProperties;
import eu.unicore.gateway.util.CustomSocketConnector;
import eu.unicore.gateway.util.LogUtil;

/**
 * Wraps a Jetty server and allows to configure it using the Gateway properties file.<br/>
 * Adds two servlets, one serving static resources for the website, the other handling 
 * requests.
 * 
 * @author schuller
 */
public class JettyServer {

	private static final Logger logger=LogUtil.getLogger(LogUtil.GATEWAY,JettyServer.class);

	private Server theServer;
	private URI url;
	private int port;
	private String host;
	private Context root;
	private SecurityProperties securityProperties;
	private GatewayProperties gatewayProperties;
	private JettyProperties jettyProperties;
	private Gateway gateway;


	public JettyServer(Gateway gateway)
	throws Exception
	{
		gatewayProperties=gateway.getProperties();
		securityProperties=gateway.getSecurityProperties();
		jettyProperties = gateway.getJettyProperties();
		this.gateway = gateway;
		initServer();
	}

	public void start() throws Exception{
		theServer.start();
	}

	public void stop() throws Exception{
		theServer.stop();
	}

	protected void initServer() throws Exception{
		System.setProperty("org.mortbay.log.class", JettyLogger.class.getName()); 
		String hostProp=gatewayProperties.getHostname();
		url=URI.create(hostProp);
		try{
			port=url.getPort();
			host=url.getHost();
		}catch(Exception e){}
		if("0.0.0.0".equals(host)){
			logger.info("Starting Jetty, will listen on all network interfaces.");
		}
		else{
			logger.info("Starting Jetty on "+host);	
		}
		theServer=new Server();
		Connector connector=null;
		if(jettyProperties.getBooleanProperty(JettyProperties.ENABLE_AJP,false)){
			connector=createAJPConnector();
		}
		else if(url.getScheme().startsWith("https")){
			connector=createSecureConnector();
		}
		else{
			connector=createPlainConnector();
		}
		theServer.addConnector(connector);
		if (jettyProperties.getBooleanProperty(JettyProperties.FAST_RANDOM, true)){
			logger.info("Using fast (but less secure) session ID generator. Set the property <"
					+ JettyProperties.FAST_RANDOM
					+ "> to 'false' to enable use of SecureRandom.");
			// use not-so-secure thingy
			SessionIdManager sm = new HashSessionIdManager(new java.util.Random());
			theServer.setSessionIdManager(sm);
		}
		configureConnector(connector);
		configureServer();
		addServlets();
	}

	protected void configureServer() {
		QueuedThreadPool btPool=new QueuedThreadPool();
		btPool.setMaxThreads(jettyProperties.getIntProperty(JettyProperties.MAX_THREADS));
		btPool.setMinThreads(jettyProperties.getIntProperty(JettyProperties.MIN_THREADS));
		btPool.setMaxIdleTimeMs(jettyProperties.getIntProperty(JettyProperties.MAX_IDLE_TIME));
		btPool.setLowThreads(jettyProperties.getIntProperty(JettyProperties.LOW_THREADS));
		if(btPool.getLowThreads()>btPool.getMaxThreads()){
			logger.debug("Resetting lowThreads parameter to '0' (must be smaller than maxThreads)");
			btPool.setLowThreads(0);
		}
		theServer.setThreadPool(btPool);
	}

	protected Connector createSecureConnector(){
		if(jettyProperties.getBooleanProperty(JettyProperties.USE_NIO, false)){
			return createNIOConnector();
		}
		logger.info("Creating SSL connector on port "+port);
		CustomSslSocketConnector ssl = new CustomSslSocketConnector(gatewayProperties);
		//configure ssl settings
		ssl.setNeedClientAuth(true);
		ssl.setKeystore(securityProperties.getKeystore());
		ssl.setKeyPassword(securityProperties.getKeystorePassword());
		ssl.setPassword(securityProperties.getKeystorePassword());
		ssl.setKeystoreType(securityProperties.getKeystoreType());
		ssl.setTruststore(securityProperties.getTruststore());
		ssl.setTruststoreType(securityProperties.getTruststoreType());
		ssl.setTrustPassword(securityProperties.getTruststorePassword());
		//fix for IBM JDK where default protocol "TLS" does not work
		String vm=System.getProperty("java.vm.vendor");
		if(vm!=null && vm.trim().startsWith("IBM")){
			ssl.setProtocol("SSL_TLS");//works for clients using both SSLv3 and TLS
			logger.info("For IBM JDK: Setting SSL protocol to '"+ssl.getProtocol()+"'");
		}
		configureConnector(ssl);
		return ssl;
	}

	protected Connector createNIOConnector(){
		logger.info("Creating NIO SSL connector on port "+port);
		NIOSSLSocketConnector ssl=new NIOSSLSocketConnector(gatewayProperties);
		ssl.setLowResourceMaxIdleTime(jettyProperties.getIntProperty(JettyProperties.LOW_RESOURCE_MAX_IDLE_TIME));
		long lowResourcesConnections=100;
		try{
			lowResourcesConnections=jettyProperties.getIntProperty(JettyProperties.MAX_THREADS)-
			jettyProperties.getIntProperty(JettyProperties.LOW_THREADS);
		}catch(Exception e){}
		ssl.setLowResourcesConnections(lowResourcesConnections);
		ssl.setNeedClientAuth(true);
		ssl.setKeystore(securityProperties.getKeystore());
		ssl.setKeyPassword(securityProperties.getKeystorePassword());
		ssl.setPassword(securityProperties.getKeystorePassword());
		ssl.setKeystoreType(securityProperties.getKeystoreType());
		ssl.setTruststore(securityProperties.getTruststore());
		ssl.setTruststoreType(securityProperties.getTruststoreType());
		ssl.setTrustPassword(securityProperties.getTruststorePassword());
		configureConnector(ssl);
		return ssl;
	} 

	protected AbstractConnector createPlainConnector(){
		logger.info("Creating plain HTTP connector.");
		AbstractConnector connector= new CustomSocketConnector(gatewayProperties);
		configureConnector(connector);
		return connector;
	}

	protected AbstractConnector createAJPConnector(){
		logger.info("Creating AJP13 connector.");
		Ajp13SocketConnector connector= new Ajp13SocketConnector();
		configureConnector(connector);
		return connector;
	}

	/**
	 * sets common parameters on the Connector, like
	 * hostname and port
	 */
	protected void configureConnector(Connector http){
		http.setHost(host);
		http.setPort(port);
		http.setLowResourceMaxIdleTime(jettyProperties.getIntProperty(JettyProperties.LOW_RESOURCE_MAX_IDLE_TIME));
	}


	protected void addServlets()throws Exception {
		root=new Context(theServer, "/", Context.SESSIONS);
		ServletHolder servletHolder = new ServletHolder(new Servlet(gateway));
		root.addServlet(servletHolder, "/*");

		boolean enableGzip=jettyProperties.isJettyGzipEnabled();
		if(enableGzip){
			//add gzip filter
			FilterHolder gzipHolder = new FilterHolder(
					new ConfigurableGzipFilter(jettyProperties));
			root.addFilter(gzipHolder, "/", Handler.REQUEST);
			logger.info("Enabling GZIP compression filter.");
		}

		URL u=getClass().getResource("/eu/unicore/gateway");
		root.setResourceBase(u.toString());
		logger.debug("Adding resources servlet, base="+u.toString());
		root.addServlet(DefaultServlet.class,"/resources/*");
	}

	public Context getRootContext(){
		return root;
	}

	/**
	 * get the Jetty server
	 * @return server
	 */
	public Server getServer(){
		return theServer;
	}


	public GatewayProperties getGatewayProperties(){
		return gatewayProperties;
	}

}
