/*********************************************************************************
 * Copyright (c) 2008 Forschungszentrum Juelich GmbH 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * (1) Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the disclaimer at the end. Redistributions in
 * binary form must reproduce the above copyright notice, this list of
 * conditions and the following disclaimer in the documentation and/or other
 * materials provided with the distribution.
 * 
 * (2) Neither the name of Forschungszentrum Juelich GmbH nor the names of its 
 * contributors may be used to endorse or promote products derived from this 
 * software without specific prior written permission.
 * 
 * DISCLAIMER
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 ********************************************************************************/

package eu.unicore.gateway.base;

import java.io.IOException;
import java.io.InputStream;
import java.net.InetSocketAddress;
import java.net.Socket;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.NoSuchProviderException;
import java.security.SecureRandom;
import java.security.cert.CertificateException;

import javax.net.ssl.HandshakeCompletedEvent;
import javax.net.ssl.HandshakeCompletedListener;
import javax.net.ssl.KeyManager;
import javax.net.ssl.KeyManagerFactory;
import javax.net.ssl.SSLContext;
import javax.net.ssl.SSLPeerUnverifiedException;
import javax.net.ssl.SSLSocket;
import javax.net.ssl.TrustManager;
import javax.net.ssl.TrustManagerFactory;
import javax.net.ssl.X509TrustManager;

import org.apache.log4j.Logger;
import org.mortbay.jetty.security.Password;
import org.mortbay.jetty.security.SslSelectChannelConnector;
import org.mortbay.resource.Resource;

import eu.unicore.gateway.properties.GatewayProperties;
import eu.unicore.gateway.util.LogUtil;
import eu.unicore.security.util.KeystoreUtil;

/**
 * customised version of NIO SSL connector, that allows to customize trust
 * (e.g. for proxy validation)
 * 
 * @author schuller
 */
public class NIOSSLSocketConnector extends SslSelectChannelConnector {
	
	private final static Logger log=LogUtil.getLogger(LogUtil.CONNECTIONS,NIOSSLSocketConnector.class);
	
	private transient Password password;
	private transient Password keyPassword;
	private transient Password trustPassword;
	private GatewayProperties properties;
	
	public NIOSSLSocketConnector(GatewayProperties properties)
	{
		this.properties = properties;
		setTruststoreType(null);
		setKeystoreType(null);
	}
	
	public void setPassword(String password){
		this.password=Password.getPassword(PASSWORD_PROPERTY,password,null);
		super.setPassword(password);
	}

	public void setTrustPassword(String password){
		trustPassword=Password.getPassword(PASSWORD_PROPERTY,password,null);
		super.setTrustPassword(password);
	}

	public void setKeyPassword(String password){
		keyPassword=Password.getPassword(KEYPASSWORD_PROPERTY,password,null);
		super.setKeyPassword(password);
	}

	protected SSLContext createSSLContext() throws Exception
	{
		if (getTruststore()==null)
		{
			setTruststore(getKeystore());
			setTruststoreType(getKeystoreType());
		}

		KeyManager[] keyManagers=null;
		InputStream keystoreInputStream = null;
		if (getKeystore()!=null)
			keystoreInputStream=Resource.newResource(getKeystore()).getInputStream();
		KeyStore keyStore=KeyStore.getInstance(getKeystoreType());
		keyStore.load(keystoreInputStream,password==null?null:password.toString().toCharArray());

		KeyManagerFactory keyManagerFactory=KeyManagerFactory.getInstance(getSslKeyManagerFactoryAlgorithm());
		keyManagerFactory.init(keyStore,keyPassword==null?null:keyPassword.toString().toCharArray());
		keyManagers=keyManagerFactory.getKeyManagers();

		KeyStore trustStore=loadTruststore();
		TrustManagerFactory trustManagerFactory=TrustManagerFactory.getInstance(getSslTrustManagerFactoryAlgorithm());
		trustManagerFactory.init(trustStore);
		TrustManager[] tm=trustManagerFactory.getTrustManagers();
		TrustManager[] trustManagers=new X509TrustManager[tm.length];
		for(int i=0;i<trustManagers.length;i++){
			trustManagers[i]=decorate(tm[i]);
		}
		
		
		SecureRandom secureRandom=getSecureRandomAlgorithm()==null?null:SecureRandom.getInstance(getSecureRandomAlgorithm());
		SSLContext context=getProvider()==null?SSLContext.getInstance(getProtocol()):SSLContext.getInstance(getProtocol(),getProvider());
		context.init(keyManagers,trustManagers,secureRandom);
		return context;
	}

	protected TrustManager decorate(TrustManager tm){
		return tm instanceof X509TrustManager ? 
			new AuthSSLX509TrustManager((X509TrustManager)tm, properties) : tm ;
	}

	protected KeyStore loadTruststore()throws IOException,NoSuchAlgorithmException,CertificateException,KeyStoreException, NoSuchProviderException {
		String name=getTruststore();
		String passwd=trustPassword.toString();
		String type=getTruststoreType();
		return KeystoreUtil.loadTruststore(name, passwd, type);
	}

	@Override
	protected void configure(Socket socket)throws IOException{
		InetSocketAddress peer=(InetSocketAddress)socket.getRemoteSocketAddress();
		if(log.isDebugEnabled() && peer!=null && peer.getAddress()!=null){
			log.debug("Connection attempt from "+peer.getAddress().getHostAddress());
			if(socket instanceof SSLSocket){
				SSLSocket ssl=(SSLSocket)socket;
				ssl.addHandshakeCompletedListener(new HandshakeCompletedListener(){

					public void handshakeCompleted(HandshakeCompletedEvent hce) {
						String msg;
						try{
							msg=String.valueOf(hce.getPeerPrincipal());
						}catch(SSLPeerUnverifiedException spe){
							msg="Problem: Unverified peer!";
						}
						log.debug("SSL connection to <"+msg+"> established.");
					}
					
				});
			}
		}
		super.configure(socket);
	}
	
	
}
