/*********************************************************************************
 * Copyright (c) 2006-2010 Forschungszentrum Juelich GmbH 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * (1) Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the disclaimer at the end. Redistributions in
 * binary form must reproduce the above copyright notice, this list of
 * conditions and the following disclaimer in the documentation and/or other
 * materials provided with the distribution.
 * 
 * (2) Neither the name of Forschungszentrum Juelich GmbH nor the names of its 
 * contributors may be used to endorse or promote products derived from this 
 * software without specific prior written permission.
 * 
 * DISCLAIMER
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 ********************************************************************************/


package eu.unicore.gateway.base;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.net.URI;
import java.net.URL;
import java.security.cert.X509Certificate;
import java.util.Enumeration;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.HttpMethod;
import org.apache.commons.httpclient.methods.GetMethod;
import org.apache.commons.httpclient.methods.InputStreamRequestEntity;
import org.apache.commons.httpclient.methods.PutMethod;
import org.apache.log4j.Logger;
import org.apache.log4j.MDC;

import eu.unicore.gateway.Gateway;
import eu.unicore.gateway.POSTHandler;
import eu.unicore.gateway.SiteOrganiser;
import eu.unicore.gateway.VSite;
import eu.unicore.gateway.client.HttpClientFactory;
import eu.unicore.gateway.properties.GatewayProperties;
import eu.unicore.gateway.util.LogUtil;

/**
 * Simple servlet for doing GET, PUT and POST<br/>
 * 
 * Does not give full HTTP support, for example it does not forward all
 * the headers in the request.
 * 
 * @author roger
 * @author schuller
 */
public class Servlet extends HttpServlet {
	
	private static final long serialVersionUID=2L;
	
	private static final Logger logger=LogUtil.getLogger(LogUtil.GATEWAY,Servlet.class);
	private GatewayProperties properties;
	private Gateway gateway;
	private final HttpClientFactory clientFactory;
	
	//header name for pushing consignor IP to vsite
	public static final String X_UNICORE_CONSIGNOR_IP="X-UNICORE-Consignor-IP";
	
	public Servlet(Gateway gw)
	{
		this.gateway = gw;
		this.properties = gw.getProperties();
		clientFactory = gateway.getClientFactory();
	}
	
	@Override
	protected void service(HttpServletRequest req, HttpServletResponse res) throws ServletException, IOException{
		String clientIP=req.getRemoteAddr();
		String clientName=null;
		try{
			X509Certificate[] certPath = (X509Certificate[]) req.getAttribute("javax.servlet.request.X509Certificate");
			if (certPath != null){
				clientName=certPath[0].getSubjectDN().getName();
			}
		}catch(Exception ex){
			LogUtil.logException("Error getting client information.", ex, logger);
		}
		MDC.put(LogUtil.MDC_IP, clientIP);
		if(clientName!=null)MDC.put(LogUtil.MDC_DN, clientName);
		
		if(logger.isDebugEnabled()){
			StringBuilder sb=new StringBuilder();
			sb.append("Processing request from ").append(clientIP);
			if(clientName!=null){
				sb.append(" Client name: ").append(clientName);
			}
			logger.debug(sb.toString());
		}
		try{
			super.service(req, res);
		}finally{
			MDC.remove(LogUtil.MDC_DN);
			MDC.remove(LogUtil.MDC_IP);
		}
	}
	
	@Override
	protected void doHead(HttpServletRequest req, HttpServletResponse resp)
			throws ServletException, IOException {
		logger.debug("Processing HEAD request");
		String update=req.getHeader("Upgrade");
		if(update!=null){
			logger.debug("Switching protocols to: "+update);
			//just set the 101 status, the CustomHttpConnection 
			//will deal further with this
			doProtocolSwitch(update, resp);
			return;
		}
		else super.doHead(req, resp);
	}

	protected void doProtocolSwitch(String protocol, HttpServletResponse resp) throws IOException{
		String clazz=properties.getProtocolPluginHandler(protocol);
		if (clazz.equals(""))
		{
			String msg="Requested protocol <"+protocol+"> is not supported";
			logger.warn(msg);
			resp.sendError(500, msg);
			return;
		}
		try{
			Class.forName(clazz).newInstance();
			resp.setHeader("Upgrade", "protocol");
			resp.setStatus(101);
			resp.flushBuffer();
		}catch(Exception e){
			String msg="Plugin class for requested protocol <"+protocol+"> not found";
			LogUtil.logException(msg, e, logger);
			resp.sendError(500,msg);
		}
	}
	
	@Override
	protected void doGet(HttpServletRequest req, HttpServletResponse res) throws ServletException, IOException
	{
		SiteOrganiser so = gateway.getSiteOrganiser();     
		String url=fullRequestURL(req);

		debugRequest("GET", req, url);
		String clientIP=req.getRemoteAddr();
		
		//check if we can map the request to a virtual site
		VSite vsite=so.match(url,clientIP);
		if(vsite!=null){
			doForwardGETtoVSite(vsite, url, req, res, clientIP);
			return;
		}
		else{
			//check if the path is "/" and throw a 404 if this is not the case
			URL u=new URL(url);
			if(!"/".equals(u.getPath())){
				res.sendError(HttpServletResponse.SC_NOT_FOUND,"Could not find the requested resource.");
				return;
			}
			//show the default "monkey page"
			doGETDefaultGWPage(req, res);
		}
	}

	/**
	 * forward the GET request to the selected VSite
	 */
	private void doForwardGETtoVSite(VSite vsite, String url, HttpServletRequest req, HttpServletResponse res, String clientIP)
	throws ServletException, IOException{
		
		OutputStream os=res.getOutputStream();
		//forward the request and copy the response to our servlet response
		URI u=URI.create(vsite.resolve(url));
		GetMethod get=new GetMethod(u.getPath());
		get.setQueryString(req.getQueryString());
		addClientIPHeader(clientIP, get);
		try
		{
			logger.debug("Forwarding <"+url+"> to <"+u.toString()+">");
			HttpClient client = clientFactory.makeHttpClient(u.toURL());
			client.executeMethod(get);
			int result=get.getStatusCode();
			
			if(result<200 || result >299 ){
				res.sendError(result, get.getStatusText());
				return;
			}
			
			res.setStatus(result);
			InputStream is=get.getResponseBodyAsStream();
			byte[] buf=new byte[1024];
			int r;
			while((r=is.read(buf))!=-1){
				os.write(buf, 0, r);
			}
			if(get.getResponseHeader("content-type")!=null){
				res.setContentType(get.getResponseHeader("content-type").getValue());
			}
			
		}catch(Exception e){
			LogUtil.logException("Error performing GET request.", e, logger);
			res.sendError(500, "Could not perform request.");
		}
		finally{
			try{get.releaseConnection();}catch(Exception e){}
		}
		os.flush();
		return;
	}
	
	/**
	 * show the default Gateway page ("monkey page")
	 */
	private void doGETDefaultGWPage(HttpServletRequest req, HttpServletResponse res)throws IOException{
		SiteOrganiser so = gateway.getSiteOrganiser();  
		PrintWriter out=res.getWriter();
		//else show the gw default page
		res.setContentType("text/html");
		X509Certificate[] certs = (X509Certificate[]) req.getAttribute("javax.servlet.request.X509Certificate");

		out.println("<html><link rel='stylesheet' type='text/css' href='resources/gateway.css'/>"+
		"<title>UNICORE 6 Gateway</title><body>");

		String top="<div id='header'><a href='http://www.unicore.eu'><img src='resources/unicore_logo.gif' border='0'/></a>" +
		"<br/> Gateway <br/>";
		if (certs != null)
		{
			top+=("<p class='username'>You are authenticated as: <br/>"+certs[0].getSubjectDN() + "</p>");
		}
		top+="</div>";

		out.println(getContentDiv(top));
		out.println("<br/>");
		if(!properties.isDetailedWebPageDisabled()){
			out.println(getContentDiv(so.toHTMLString()));
		}
		else{
			out.println(getContentDiv("<br/>Detailed site listing disabled.<br/>"));
		}
		out.println("<br/>");

		out.println(getFooter());

		out.println("</html></body>");
	}
	
	@Override
	protected void doPut(HttpServletRequest req, HttpServletResponse res) throws ServletException, IOException{
		SiteOrganiser so = gateway.getSiteOrganiser();     
		String url=fullRequestURL(req);
		String clientIP=req.getRemoteAddr();
		debugRequest("PUT", req, url);
		
		//check if we can map the request to a virtual site
		VSite vsite=so.match(url,clientIP);
		if(vsite!=null){
			//forward the request and copy the response to our servlet response
			URI u=URI.create(vsite.resolve(url));
			PutMethod put=new PutMethod(u.getPath());
			put.setQueryString(req.getQueryString());
			put.setRequestEntity(new InputStreamRequestEntity(req.getInputStream()));
			put.setContentChunked(true);
			addClientIPHeader(clientIP, put);
			try
			{
				logger.debug("Forwarding <"+url+"> to <"+u.toString()+">");
				HttpClient client = clientFactory.makeHttpClient(u.toURL());
				client.executeMethod(put);
				int result=put.getStatusCode();
				if(result<200 || result >299 ){
					res.sendError(result, put.getStatusText());
					return;
				}
				
				res.setStatus(result);
				InputStream is=put.getResponseBodyAsStream();
				if(is!=null){
					OutputStream bos=res.getOutputStream();
					byte[] buf=new byte[1024];
					int r;
					while((r=is.read(buf))!=-1){
						bos.write(buf, 0, r);
					}
					bos.flush();
					if(put.getResponseHeader("content-type")!=null){
						res.setContentType(put.getResponseHeader("content-type").getValue());
					}
				}
			}catch(Exception e){
				LogUtil.logException("Error performing PUT request.", e, logger);
				res.sendError(500, "Could not perform request.");
			}
			finally{
				try{put.releaseConnection();}catch(Exception e){}
			}
			return;
		}

	}


	@Override
	protected void doPost(HttpServletRequest req, HttpServletResponse res) throws ServletException, IOException
	{
		URL url=new URL(fullRequestURL(req));
		debugRequest("POST", req, url.toString());
		
		try
		{
			//check if it is a registration
			if("/VSITE_REGISTRATION_REQUEST".equals(url.getPath())){
				handleRegistration(req,res);
			}
			else{
				int maxHeaderSize = properties.getMaxSoapHeader();
				RawMessageExchange exchange = new RawMessageExchange(req.getReader(), 
						res.getWriter(), maxHeaderSize);
				String contentType = req.getHeader("Content-type");
				exchange.setProperty(RawMessageExchange.CONTENT_TYPE, contentType);
				if(contentType!=null){
					if(contentType.contains("application/soap+xml")){
						res.setContentType("application/soap+xml; charset=UTF-8");
					}
					else if(contentType.contains("application")){
						res.setContentType("application/xml; charset=UTF-8");
					}
					else{
						res.setContentType("text/xml; charset=UTF-8");
					}
				}
				else{
					res.setContentType("text/xml; charset=UTF-8");
				}
				X509Certificate[] certPath = (X509Certificate[]) req.getAttribute("javax.servlet.request.X509Certificate");
				if (certPath != null)
				{
					exchange.setProperty(RawMessageExchange.X509, certPath);
				}
				exchange.setProperty(RawMessageExchange.REMOTE_IP, req.getRemoteHost());
				
				exchange.setRequestURL(url.toString());
				exchange.setProperty(RawMessageExchange.SOAP_ACTION, req.getHeader("SOAPAction"));
				
				POSTHandler handler = new POSTHandler(gateway.getSiteOrganiser(), 
					gateway.getConsignorProducer(), clientFactory,
					properties.isChunkedDispatch(), 
					gateway.getHostURI().toString());
				handler.invoke(exchange);
				res.flushBuffer();
			}
		}
		catch (Exception e){
			LogUtil.logException("Failed to process POST request", e, logger);
			throw new ServletException("Failed to process POST request: " + e, e);
		}
	}

	protected void handleRegistration(HttpServletRequest req, HttpServletResponse res)throws ServletException,IOException{
		if(!properties.isDynamicRegistrationEnabled()){
			res.sendError(HttpServletResponse.SC_FORBIDDEN,"Dynamic registration is disabled.");
		}
		else{
			String name=req.getParameter("name");
			String address=req.getParameter("address");
			boolean success=true;
			try{
				URI uri=new URI(address);
				success=gateway.getDynamicSiteOrganiser().register(name,uri);
				if(success){
					PrintWriter pw=res.getWriter();
					pw.write("<html><body>Your request was processed successfully. <br/>" +
							"Please click <a href='/'>here</a> to return to the main gateway page.");
					res.setStatus(HttpServletResponse.SC_CREATED);
				}
			}catch(Exception e){
				LogUtil.logException("Error on registration of [name="+name+" address="+address+"]", e, logger);
			}
			if(!success){
				res.sendError(HttpServletResponse.SC_BAD_REQUEST,"Registration could not be processed.");
			}
		}
		res.flushBuffer();
	}
	
	/**
	 * reconstructs and returns the full request URL
	 * @param req - a {@link HttpServletRequest}
	 */
	private String fullRequestURL(HttpServletRequest req){
		String query=req.getQueryString();
		StringBuffer requestURL=req.getRequestURL();
		if(query!=null){
			requestURL.append("?");
			requestURL.append(query);
		}
		return requestURL.toString();
	}
	
	private String getContentDiv(String content){
		String s="<div id='content'><b class='rtop'><b class='r1'></b><b class='r2'>"+
		"</b> <b class='r3'></b> <b class='r4'></b></b>"+content+
		"<b class=<'rbottom'><b class='r4'></b> <b class='r3'></b> <b class='r2'></b> <b class='r1'></b></b></div>";
		return s;
	}

	private String getFooter(){
		StringBuilder sb=new StringBuilder();
		sb.append("<div id='footer'><hr/> Version: "+Gateway.RELEASE_VERSION+" Up since: ").append(gateway.upSince());
		if(properties.isDynamicRegistrationEnabled()){
			sb.append("&nbsp;&nbsp;&nbsp;&nbsp;<a href='resources/register.html'>register a site</a>");
		}
		sb.append("</div>");
		return sb.toString();
	}

	private void debugRequest(String type, HttpServletRequest req, String url)
	{
		if(!logger.isDebugEnabled())
			return;
		logger.debug("New " + type + " message to " + url);
		if (!logger.isTraceEnabled())
			return;

		Enumeration<?> hdrNames = req.getHeaderNames();
		StringBuilder hdrDump = new StringBuilder();
		
		while(hdrNames.hasMoreElements())
		{
			String n = (String)hdrNames.nextElement();
			hdrDump.append(n + ": " + req.getHeader(n) + "\n");
		}
		logger.trace(type+" request header:\n" + hdrDump);
	}
	
	/**
	 * adds a header X-UNICORE-Consignor-IP which is set to the given IP
	 * @param ip - the consignor IP
	 * @param method
	 */
	public static void addClientIPHeader(String ip, HttpMethod method){
		if(ip!=null){
			if(logger.isDebugEnabled()){ 
				logger.debug("Generating "+X_UNICORE_CONSIGNOR_IP+" header: "+ip);
			}
			method.setRequestHeader(X_UNICORE_CONSIGNOR_IP, ip);
		}
	}
	
}
