/**
 * Copyright (c) 2005, Forschungszentrum Juelich
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification, are
 * permitted provided that the following conditions are met: 
 * 
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *  * Neither the name of the Forschungszentrum Juelich nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package eu.unicore.gateway.client;

import java.io.File;
import java.net.URL;

import org.apache.commons.httpclient.HostConfiguration;
import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.HttpConnectionManager;
import org.apache.commons.httpclient.HttpsURL;
import org.apache.commons.httpclient.MultiThreadedHttpConnectionManager;
import org.apache.commons.httpclient.methods.PostMethod;
import org.apache.commons.httpclient.methods.RequestEntity;
import org.apache.commons.httpclient.params.HttpClientParams;
import org.apache.commons.httpclient.params.HttpConnectionManagerParams;
import org.apache.commons.httpclient.protocol.Protocol;
import org.apache.commons.httpclient.protocol.ProtocolSocketFactory;
import org.apache.log4j.Logger;

import eu.unicore.gateway.properties.GatewayProperties;
import eu.unicore.gateway.properties.SecurityProperties;
import eu.unicore.gateway.util.LogUtil;
import eu.unicore.security.util.client.CustomHostConfiguration;
import eu.unicore.security.util.client.HttpHostFactory;

public class HttpClientFactory
{
	private static final Logger log = LogUtil.getLogger(LogUtil.GATEWAY,HttpClientFactory.class);

	private int sockettimeout;
	private int connectiontimeout;
	private boolean keepAlive;
	private boolean useExpectContinue;
	private boolean enableGzip;
	private boolean chunked;
	private int maxTotalConn;
	private int maxConnPerHost;
	
	private final AuthSSLProtocolSocketFactory aspsf;
	private final HttpConnectionManager manager;
	
	public HttpClientFactory(SecurityProperties securityprops, GatewayProperties props) throws Exception{
		String keystore = new File(securityprops.getKeystore()).getAbsolutePath();
		log.debug("using keystore :" + keystore);
		String truststore = new File(securityprops.getTruststore()).getAbsolutePath();
		log.debug("using truststore :" + truststore);

		this.sockettimeout = props.getSocketTimeout();
		this.connectiontimeout = props.getConnectionTimeout();
		this.keepAlive=props.isKeepAlive();
		this.enableGzip=props.isGzipEnabled();
		this.chunked=props.isChunkedDispatch();
		this.useExpectContinue=props.isExpectContinueEnabled();
		this.maxTotalConn = props.getMaxTotalConnections();
		this.maxConnPerHost = props.getMaxPerServiceConnections();
		
		aspsf=new AuthSSLProtocolSocketFactory(securityprops);
		manager=getHttpConnectionManager();
	}

	public HttpClient makeHttpClient(URL url) throws Exception{
		HttpClient client = new HttpClient();
		client.getHostConfiguration().setHost(url.getHost(), url.getPort());
		client.getParams().setParameter(HttpClientParams.SO_TIMEOUT, sockettimeout);
		//client.getHttpConnectionManager().getParams().setLinger(100);
		client.getParams().setParameter(HttpClientParams.CONNECTION_MANAGER_TIMEOUT, (long)connectiontimeout);
		client.getParams().setParameter(HttpClientParams.USE_EXPECT_CONTINUE, useExpectContinue);
		
		if (url.getProtocol().toLowerCase().equals("https"))
		{
			Protocol authhttps = new Protocol("https", aspsf, url.getPort());
			client.getHostConfiguration().setHost(url.getHost(), url.getPort(), authhttps);
		}
		else if (url.getProtocol().toLowerCase().equals("http"))
		{
			client.getHostConfiguration().setHost(url.getHost(), url.getPort());
		}
		client.setHttpConnectionManager(manager);
		
		return client;
	}

	
	private HttpConnectionManager getHttpConnectionManager(){
		MultiThreadedHttpConnectionManager manager = new MultiThreadedHttpConnectionManager();
		HttpConnectionManagerParams conParams = new HttpConnectionManagerParams (); 
		conParams.setDefaultMaxConnectionsPerHost(maxConnPerHost);
		conParams.setMaxTotalConnections(maxTotalConn);
		conParams.setSoTimeout(sockettimeout);
		manager.setParams(conParams);
		return manager;
	}
	
	public PostMethod makePostMethod(String path, RequestEntity requestentity){
		PostMethod post = new PostMethod(path);
		post.setRequestEntity(requestentity);
		post.setRequestHeader("Content-type", "text/xml; charset=ISO-8859-1");
		if(enableGzip)post.setRequestHeader("Accept-Encoding", "gzip");
		if(!keepAlive)post.setRequestHeader("Connection", "close");
		post.setContentChunked(chunked);
		return post;
	}

	public static HostConfiguration getHostConfiguration(SecurityProperties sec)throws Exception{
		AuthSSLProtocolSocketFactory protocolFact=new AuthSSLProtocolSocketFactory(sec);
		HttpHostFactory fact=new HttpHostFactory(null,
				new Protocol(new String(HttpsURL.DEFAULT_SCHEME),
						(ProtocolSocketFactory)protocolFact,
						HttpsURL.DEFAULT_PORT));
		HostConfiguration c=new CustomHostConfiguration(fact); 
		return c;
	}

	public AuthSSLProtocolSocketFactory getSSLSocketFactory(){
		return aspsf;
	}
}
