/**
 * Copyright (c) 2005, Forschungszentrum Juelich
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification, are
 * permitted provided that the following conditions are met: 
 * 
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *  * Neither the name of the Forschungszentrum Juelich nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package eu.unicore.gateway.properties;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.InputStream;
import java.util.Properties;
import java.util.Set;

import org.apache.log4j.Logger;

/**
 * Generic properties handling.
 * @author roger
 * @author golbi
 */
public abstract class RProperties
{
	protected Properties properties;
	private Set<String> mandatory;

	/**
	 * @param name resource name to be read by class loader.
	 * @param mandatoryProperties set of mandatory properties names
	 */
	public RProperties(String name, Set<String> mandatoryProperties)
	{    
		mandatory = mandatoryProperties;
		loadProperties(name);
	}

	/**
	 * 
	 * @param f file name to be read by file IO code.
	 * @param mandatoryProperties set of mandatory properties names
	 */
	public RProperties(File f, Set<String> mandatoryProperties) 
	{    
		mandatory = mandatoryProperties;
		loadProperties(f);
	}


	protected String getStringProperty(String key)
	{
		return properties.getProperty(key);
	}

	protected String getStringProperty(String key, String defaultValue)
	{
		String ret = properties.getProperty(key);
		if (ret == null)
			return defaultValue;
		return ret;
	}

	/**
	 * @return a boolean property which is one of the required properties (i.e. is defined)
	 * 
	 */
	protected boolean getPropertyAsBool(String key)
	{
		if (!mandatory.contains(key))
			throw new RuntimeException("Tried to get " + key + " property as " +
					"a required property, while it is not in the required set.");
		String flag = properties.getProperty(key);
		return checkBool(key, flag);
	}

	protected boolean getPropertyAsBool(String key, boolean defaultVal)
	{
		String valS = properties.getProperty(key.toString());
		if (valS == null)
			return defaultVal;
		return checkBool(key, valS);
	}

	private boolean checkBool(String key, String valS)
	{
		if (valS.equalsIgnoreCase("true")
			|| valS.equalsIgnoreCase("yes"))
			return true;
		if (valS.equalsIgnoreCase("false")
			|| valS.equalsIgnoreCase("no"))
			return false;
		throw new RuntimeException(key.toString() + " parameter value "
			+ "is invalid. May be true|false|yes|no");
	}
	
	protected int getPropertyAsInt(String key)
	{
		if (!mandatory.contains(key))
			throw new RuntimeException("Tried to get " + key + " property as " +
					"a required property, while it is not in the required set.");
		String flag = properties.getProperty(key);    
		return Integer.parseInt(flag);
	}

	protected int getPropertyAsInt(String key, int defaultValue)
	{
		String flag = properties.getProperty(key);
		if (flag == null)
			return defaultValue;
		return Integer.parseInt(flag);
	}

	protected int getPropertyAsInt(String key, int min, int max, int defaultVal, Logger log)
	{
		String valS = properties.getProperty(key.toString());
		if (valS == null)
			return defaultVal;
		int retVal;
		try
		{
			retVal = Integer.parseInt(valS);
			if (retVal < min)
			{
				log.warn(key.toString() + " parameter value "
					+ "is too small, using default: "
					+ defaultVal);
				return defaultVal;
			}
			if (retVal > max)
			{
				log.warn(key.toString() + " parameter value "
					+ "is too big, using default: "
					+ defaultVal);
				return defaultVal;
			}
			return retVal;
		} catch (NumberFormatException e)
		{
			log.warn(key.toString() + " parameter value "
				+ "is invalid, using default: " + defaultVal);
			return defaultVal;
		}

	}

	
	private void cacheProperties(InputStream is)
	{
		properties = new Properties();
		try
		{
			properties.load(is);
		} catch (Exception e)
		{
			throw new Error("cannot find properties file.", e);
		}
		try	{
			is.close();
		}
		catch(Exception ignored) {}
		
		String propscheck = checkMandatoryProperties();
		if (propscheck.length() != 0)
		{
			throw new RuntimeException("Cannot find all necessary properties :: " 
				+ propscheck);
		}
	}


	protected void clearCache()
	{
		properties.clear();
	}

	protected void loadProperties(File f)
	{
		try
		{
			cacheProperties(new FileInputStream(f));
		} catch (FileNotFoundException e)
		{
			throw new RuntimeException("Problem finding properties file " + f.getAbsolutePath(), e);
		}		
	}

	protected void loadProperties(String name)
	{
		cacheProperties(getClass().getResourceAsStream(name));
	}
	
	private String checkMandatoryProperties()
	{
		StringBuilder builder = new StringBuilder();
		if (mandatory != null)
		{
			for (String o : mandatory)
			{
				if (properties.get(o) == null)
					builder.append(o).append(" ");
			}
		}
		return builder.toString();
	}
}




