package eu.unicore.gateway.util;

import org.apache.log4j.Logger;

public class LogUtil {
	
	protected LogUtil(){}
	
	/**
	 * logger prefix for general Gateway code
	 */
	public static final String GATEWAY="unicore.gateway";
	
	/**
	 * logger prefix for connection problem debugging
	 */
	public static final String CONNECTIONS="unicore.gateway.connections";
	
	/**
	 * logger prefix for security code
	 */
	public static final String SECURITY="unicore.gateway.security";
	
	/**
	 * logger prefix for detailed HTTP processing code
	 */
	public static final String JETTY="unicore.gateway.jetty";
	
	/**
	 * MDC context key for the client's IP
	 */
	public static final String MDC_IP="clientIP";
	
	/**
	 * MDC context key for the client's DN
	 */
	public static final String MDC_DN="clientName";
	
	
	/**
	 * returns a logger name, using the given prefix and the simple name
	 * of the given class
	 * 
	 * @param prefix - the prefix to use
	 * @param clazz - the class
	 * @return logger name
	 */
	public static String getLoggerName(String prefix, Class<?>clazz){
		return prefix+"."+clazz.getSimpleName();
	}
	
	/**
	 * returns a logger, using the given prefix and the simple name
	 * of the given class
	 * 
	 * @param prefix - the prefix to use
	 * @param clazz - the class
	 * @return logger
	 */
	public static Logger getLogger(String prefix, Class<?>clazz){
		return Logger.getLogger(prefix+"."+clazz.getSimpleName());
	}
	
	/** 
	 * log an error message to the default logger ("unicore.wsrflite")
	 * A human-friendly message is constructed and logged at "INFO" level.
	 * The stack trace is logged at "DEBUG" level.
	 * 
	 * @param message - the error message
	 * @param cause - the cause of the error
	 *
	 */
	public static void logException(String message, Throwable cause){
		logException(message,cause,Logger.getLogger(GATEWAY));
	}
	
	/**
	 * log an error message to the specified logger.
	 * A human-friendly message is constructed and logged at "INFO" level.
	 * The stack trace is logged at "DEBUG" level.
	 * 
	 * @param message - the error message
	 * @param cause - the cause of the error
	 * @param logger - the logger to use
	 */
	public static void logException(String message, Throwable cause, Logger logger){
		logger.error(message);
		if(cause!=null){
			logger.error("The root error was: "+getRootCause(cause));
			if(logger.isDebugEnabled())logger.debug("Stack trace",cause);
			else{
				logger.error("To see the full error stack trace, set log4j.logger."+logger.getName()+"=DEBUG");
			}
		}
	}
	
	/**
	 * construct a (hopefully) useful error message from the root cause of an 
	 * exception
	 * @param throwable
	 * @return the cause of the exception
	 */
	public static String getRootCause(Throwable throwable){
		StringBuilder sb=new StringBuilder();
		Throwable cause=throwable;
		String message=null;
		String type=cause.getClass().getName();
		do{
			type=cause.getClass().getName();
			message=cause.getMessage();
			cause=cause.getCause();
		}
		while(cause!=null);
		
		if(message!=null)sb.append(type).append(": ").append(message);
		else sb.append(type).append(" (no further message available)");
		return sb.toString();
	}
	
}
