/**
 * Copyright (c) 2005, Forschungszentrum Juelich
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification, are
 * permitted provided that the following conditions are met: 
 * 
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *  * Neither the name of the Forschungszentrum Juelich nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


package eu.unicore.gateway.util;

import java.net.URI;
import java.util.LinkedList;
import java.util.StringTokenizer;

/**
 * sort of adds some new things to java.net.URI - trailing slashes seem to cause me the
 * most confusion. should be seen as a drop-in replacement for java.net.URI
 * 
 * @author roger
 */
public class XURI
{

  private final URI uri;

  private URI doubledots = URI.create("..");

  private LinkedList<String> pl;

  public XURI(URI uri)
  {
    this.uri = uri.normalize();
    init();
  }
  
  public XURI(String uri)
  {
    this.uri = URI.create(uri).normalize();
    init();
  }

  public XURI(XURI u, URI path)
  {
    this.uri = u.resolve(path.normalize()).getURI();
    init();
  }
  
  public static XURI create(String uri)
  {
    return new XURI(URI.create(uri));
  }
  
  private void init()
  {
    pl = new LinkedList<String>();
    StringTokenizer tok = new StringTokenizer(getPath(), "/");
    while (tok.hasMoreTokens())
    {
      pl.add(tok.nextToken());
    }
  }

  public XURI resolve(URI ruri)
  {
    return new XURI(getURI().resolve(ruri));
  }

  public URI getSchemeAndAuthority()
  {
    return URI.create(getURI().getScheme() + "://" + getURI().getAuthority());
  }

  public String getPath()
  {
    return getURI().getPath();
  }

  public String getHost()
  {
    return getURI().getHost();
  }
  
  public int getPort()
  {
    return getURI().getPort();
  }
  
  public String getScheme()
  {
    return getURI().getScheme();
  }
  
  
  /**
   * resolving "/a/b/c" with ".." gives "/a/", but ... resolving "/a/b/c/" with ".." gives
   * "/a/b/". well, it does kind of make sense, but you still have to watch out for it
   * .... !
   * 
   * @return
   */
  public boolean isDirectory()
  {
    return getPath().endsWith("/");
  }
  
  public XURI ensureIsDirectory()
  {
    if (this.isDirectory())
    {
      return this;
    }
    return new XURI(URI.create(getURI()+"/"));
  }

  public XURI getParent()
  {
    if (isDirectory())
    {
      return this.resolve(doubledots);
    }
    else
    {
      return new XURI(URI.create(getURI().toString() + "/")).resolve(doubledots);
    }
  }

  public int countPathElements()
  {
    if (this.pl.size() == 1 && this.pl.getFirst().equals(".."))
    {
      return -1;
    }
    return this.pl.size();
  }

  public LinkedList<String> getPathElements()
  {
    return this.pl;
  }

  public String getPathElement(int i)
  {
    if (this.countPathElements() > i)
    {
      return getPathElements().listIterator(i).next();
    }
    else return null;
  }

  public XURI relativize(XURI xuri)
  {
    return new XURI(getURI().relativize(xuri.getURI()));
  }
  
  public URI getURI()
  {
    return this.uri;
  }

  @Override
  public boolean equals(Object o)
  {
	if (o==null || o.getClass()!=this.getClass()) return false;
    return getURI().equals(((XURI) o).getURI());
  }
  
  @Override
  public int hashCode()
  {
    return getURI().hashCode();
  }
  
  @Override
  public String toString()
  {
    return getURI().toString();
  }

}
