package eu.unicore.gateway;

import java.io.BufferedReader;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.StringReader;
import java.net.URL;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import javax.servlet.http.HttpServletRequest;
import javax.xml.namespace.QName;
import javax.xml.stream.XMLEventReader;
import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.events.XMLEvent;

import junit.framework.TestCase;

import org.apache.commons.httpclient.Header;
import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.HttpStatus;
import org.apache.commons.httpclient.methods.ByteArrayRequestEntity;
import org.apache.commons.httpclient.methods.GetMethod;
import org.apache.commons.httpclient.methods.PostMethod;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;

import eu.unicore.bugsreporter.annotation.FunctionalTest;
import eu.unicore.bugsreporter.annotation.RegressionTest;
import eu.unicore.gateway.base.Servlet;
import eu.unicore.gateway.properties.GatewayProperties;
import eu.unicore.gateway.soap.Soap11;
import eu.unicore.gateway.soap.Soap12;
import eu.unicore.gateway.soap.SoapVersion;
import eu.unicore.gateway.util.ProtocolPluginHandler;
import eu.unicore.gateway.util.UpdateProtocolMethod;

public class TestServer extends TestCase {
	static Gateway gw;
	static boolean haveInit=false;
	
	@Override
	protected void setUp() throws Exception{
		File gp = new File("src/test/resources/gateway.properties");
		File sp = new File("src/test/resources/security.properties");
		File cp = new File("src/test/resources/connection.properties");
		gw = new Gateway(gp, cp, sp);
		gw.startGateway();
	}
	
	protected void tearDown()throws Exception{
		Thread.sleep(2000);
		gw.stopGateway();
	}
	public void testSwitchProtocols()throws Exception{
		URL url=new URL("http://localhost:64433/foo");
		HttpClient hc = gw.getClientFactory().makeHttpClient(url);
		UpdateProtocolMethod post=new UpdateProtocolMethod(new ProtocolPluginHandler(){
			public void handle(InputStream is, OutputStream os)throws IOException{
				BufferedReader br=new BufferedReader(new InputStreamReader(is));
				byte[] data="Testdata!!\n".getBytes();
				for(int i=0;i<2;i++){
					System.out.print("send : "+new String(data));
					os.write(data);
					os.flush();
					String line=br.readLine();
					if(line!=null){
						System.out.println("recv : "+line);
						assertTrue(line.contains("Testdata"));
					}
				}
				is.close();
				os.close();
			}
			public String getProtocol(){return "ECHO";}
			public void configure(HttpServletRequest req){}
		});
		
		hc.executeMethod(post);
		assertTrue("Error occured, server replied: "+post.getStatusLine(),post.getStatusCode()==101);
		post.releaseConnection();
	}
	
	@FunctionalTest(id="gw_dynReg", description="Tests whether dynamic registration is possible and if can be disabled.")
	public void testDynamicRegistration()throws Exception{
		gw.getProperties().setProperty(GatewayProperties.KEY_REG_ENABLED, "false");
		int status=doRegister("test","http://localhost:12345");
		assertEquals(HttpStatus.SC_FORBIDDEN,status);
		gw.getProperties().setProperty(GatewayProperties.KEY_REG_ENABLED, "true");
		status=doRegister("test","http://localhost:12345");
		assertEquals(HttpStatus.SC_CREATED,status);
	}
	
	public void testPost()throws Exception{
		FakeServer s1=new FakeServer();
		s1.start();
		String s1Url=s1.getURI();
		int status=doRegister("FAKE1",s1Url);
		assertEquals(HttpStatus.SC_CREATED,status);
		
		String url="http://localhost:64433/FAKE1/test";
		HttpClient hc = gw.getClientFactory().makeHttpClient(new URL(url));
		PostMethod post=new PostMethod(url);
		post.setRequestEntity(new ByteArrayRequestEntity(getBody(url)));
		try{
			hc.executeMethod(post);
			System.out.println(post.getStatusLine());
			status=post.getStatusCode();
			assertEquals(HttpStatus.SC_OK, status);
			assertTrue(post.getResponseHeader("Content-type").getValue().contains("text/xml"));
			assertTrue("missing client IP header",containsConsignorIP(s1.getLatestRequestHeaders()));
		}
		catch(Exception ex){
			ex.printStackTrace();
			fail();
		}
		
		int expectedCode=503;
		
		s1.setStatusCode(expectedCode);
		
		post=new PostMethod(url);
		post.setRequestEntity(new ByteArrayRequestEntity(getBody(url)));
		try{
			hc.executeMethod(post);
			System.out.println(post.getStatusLine());
			status=post.getStatusCode();
		    //TODO what is the expected behaviour? Currently the GW always returns 200 OK even
			//if the vsite gives an error code 
			//assertEquals(expectedCode, status);
		}
		catch(Exception ex){
			ex.printStackTrace();
			fail();
		}
		
		
		s1.stop();
	}
	
	@FunctionalTest(id="gw_fwPostReq", 
		description="Tests whether the gateway properly forwards POST/SOAP requests to the backend site; "
		+ "requests are of application/xml mimetype.")
	public void testPostSOAP11Content()throws Exception{
		FakeServer s1=new FakeServer();
		s1.start();
		String s1Url=s1.getURI();
		int status=doRegister("FAKE1",s1Url);
		assertEquals(HttpStatus.SC_CREATED,status);
		
		String url="http://localhost:64433/FAKE1/test";
		HttpClient hc = gw.getClientFactory().makeHttpClient(new URL(url));
		PostMethod post=new PostMethod(url);
		byte[] originalRequestBody = getBody(url);
		post.setRequestEntity(new ByteArrayRequestEntity(originalRequestBody));
		post.setRequestHeader("Content-type", "application/xml; charset=UTF-8");

		try{
			hc.executeMethod(post);
			System.out.println(post.getStatusLine());
			status=post.getStatusCode();
			assertEquals(HttpStatus.SC_OK, status);
			assertTrue(post.getResponseHeader("Content-type").getValue().contains("application/xml"));

			String forwardedRequestBody = s1.getLatestRequestBody();
			byte[] responseBody = post.getResponseBody();
			assertTrue("Server response was not properly forwarded",
					Arrays.equals(s1.getAnswer(), responseBody));

			Set<String> forwardedSoapHeaders = new HashSet<String>();
            String forwardedSoapBody = parseSoap(forwardedRequestBody, forwardedSoapHeaders);
			Set<String> originalSoapHeaders = new HashSet<String>();
            String originalSoapBody = parseSoap(new String(originalRequestBody,"ISO-8859-1"), originalSoapHeaders);

			assertEquals("SOAP body was not properly forwarded", originalSoapBody, forwardedSoapBody);
			assertTrue("Some SOAP headers were not forwarded", forwardedSoapHeaders.containsAll(originalSoapHeaders));
		}
		catch(Exception ex){
			ex.printStackTrace();
			fail();
		}
		
		int expectedCode=503;
		
		s1.setStatusCode(expectedCode);
		
		post=new PostMethod(url);
		originalRequestBody = getBody(url);
		post.setRequestEntity(new ByteArrayRequestEntity(originalRequestBody));
		try{
			hc.executeMethod(post);
			System.out.println(post.getStatusLine());
			status=post.getStatusCode();
		    //TODO what is the expected behaviour? Currently the GW always returns 200 OK even
			//if the vsite gives an error code 
			//assertEquals(expectedCode, status);
		}
		catch(Exception ex){
			ex.printStackTrace();
			fail();
		}
		
		
		s1.stop();
	}

	@FunctionalTest(id="gw_fwPostReq", 
		description="Tests whether the gateway properly forwards POST/SOAP requests to the backend site; "
		+ "requests are of text/xml mimetype.")
	public void testPostSOAP11ContentText()throws Exception{
		FakeServer s1=new FakeServer();
		s1.start();
		String s1Url=s1.getURI();
		int status=doRegister("FAKE1",s1Url);
		assertEquals(HttpStatus.SC_CREATED,status);
		
		String url="http://localhost:64433/FAKE1/test";
		HttpClient hc = gw.getClientFactory().makeHttpClient(new URL(url));
		PostMethod post=new PostMethod(url);
		byte[] originalRequestBody = getBody(url);
		post.setRequestEntity(new ByteArrayRequestEntity(originalRequestBody));
		post.setRequestHeader("Content-type", "text/xml; charset=UTF-8");

		try{
			hc.executeMethod(post);
			System.out.println(post.getStatusLine());
			status=post.getStatusCode();
			assertEquals(HttpStatus.SC_OK, status);
			assertTrue(post.getResponseHeader("Content-type").getValue().contains("text/xml"));
			
			String forwardedRequestBody = s1.getLatestRequestBody();
			byte[] responseBody = post.getResponseBody();
			assertTrue("Server response was not properly forwarded",
					Arrays.equals(s1.getAnswer(), responseBody));

			Set<String> forwardedSoapHeaders = new HashSet<String>();
            String forwardedSoapBody = parseSoap(forwardedRequestBody, forwardedSoapHeaders);
			Set<String> originalSoapHeaders = new HashSet<String>();
            String originalSoapBody = parseSoap(new String(originalRequestBody,"ISO-8859-1"), originalSoapHeaders);
			
			assertEquals("SOAP body was not properly forwarded", originalSoapBody, forwardedSoapBody);
			assertTrue("Some SOAP headers were not forwarded", forwardedSoapHeaders.containsAll(originalSoapHeaders));
		}
		catch(Exception ex){
			ex.printStackTrace();
			fail();
		}
		
		int expectedCode=503;
		
		s1.setStatusCode(expectedCode);
		
		post=new PostMethod(url);
		post.setRequestEntity(new ByteArrayRequestEntity(getBody(url)));
		try{
			hc.executeMethod(post);
			System.out.println(post.getStatusLine());
			status=post.getStatusCode();
		    //TODO what is the expected behaviour? Currently the GW always returns 200 OK even
			//if the vsite gives an error code 
			//assertEquals(expectedCode, status);
		}
		catch(Exception ex){
			ex.printStackTrace();
			fail();
		}
		
		
		s1.stop();
	}
	
	public void testPostNoSOAP()throws Exception{
		Logger.getLogger("unicore.gateway.Servlet").setLevel(Level.DEBUG);
		FakeServer s1=new FakeServer();
		s1.start();
		String s1Url=s1.getURI();
		int status=doRegister("FAKE1",s1Url);
		assertEquals(HttpStatus.SC_CREATED,status);
		
		String url="http://localhost:64433/FAKE1/test";
		HttpClient hc = gw.getClientFactory().makeHttpClient(new URL(url));
		PostMethod post=new PostMethod(url);
		post.setRequestEntity(new ByteArrayRequestEntity(getBodyNoSOAP(url)));
		try{
			hc.executeMethod(post);
			System.out.println(post.getStatusLine());
			status=post.getStatusCode();
			assertEquals(HttpStatus.SC_INTERNAL_SERVER_ERROR, status);
		}
		catch(Exception ex){
			ex.printStackTrace();
			fail();
		}
		s1.stop();
	}

	@FunctionalTest(id="gw_fwGetReq", description="Tests whether the gateway properly forwards GET requests to the backend site.")
	public void testGet()throws Exception{
		FakeServer s1=new FakeServer();
		s1.start();
		String s1Url=s1.getURI();
		int status=doRegister("FAKE1",s1Url);
		assertEquals(HttpStatus.SC_CREATED,status);

		String queryPath = "/test";
		String url="http://localhost:64433/FAKE1"+queryPath;
		String originalQuery = "GET "+queryPath+" HTTP/1.1";

		HttpClient hc = gw.getClientFactory().makeHttpClient(new URL(url));
		GetMethod get=new GetMethod(url);
		try{
			hc.executeMethod(get);
			List<String> filteredHeadersOriginal=new ArrayList<String>();
			for(Header h: get.getRequestHeaders()){
				String string=h.toExternalForm().trim();
				if(!string.startsWith("Host:")){
					filteredHeadersOriginal.add(string);
				}
			}
			List<String> filteredHeadersForwarded=new ArrayList<String>();
			for(String h: s1.getLatestRequestHeaders()){
				String string=h.trim();
				if(!string.startsWith("Host:") && !string.startsWith("X-UNICORE")){
					filteredHeadersForwarded.add(string);
				}
			}

			assertEquals("Client query was not properly forwarded", originalQuery, s1.getLatestQuery());
			assertTrue("Headers were added or modified by gateway", filteredHeadersOriginal.containsAll(filteredHeadersForwarded));
			assertTrue("Headers were not forwarded", filteredHeadersForwarded.containsAll(filteredHeadersOriginal));
			assertTrue("Missing client IP header", containsConsignorIP(s1.getLatestRequestHeaders()));
			System.out.println(get.getStatusLine());
			status=get.getStatusCode();
			assertEquals(HttpStatus.SC_OK, status);
			assertTrue("Server response was not properly forwarded",
					Arrays.equals(s1.getAnswer(), get.getResponseBody()));
		}
		catch(Exception ex){
			ex.printStackTrace();
			fail();
		}
		
		int expectedCode=503;
		s1.setStatusCode(expectedCode);
		
		get=new GetMethod(url);
		try{
			hc.executeMethod(get);
			System.out.println(get.getStatusLine());
			status=get.getStatusCode();
			assertEquals(expectedCode, status);
		}
		catch(Exception ex){
			ex.printStackTrace();
			fail();
		}
		
		
		s1.stop();
	}
	
	@RegressionTest(id=3314648, url="http://sourceforge.net/tracker/?func=detail&aid=3314648&group_id=102081&atid=633902")
	public void testGetWrongAddress()throws Exception{
		
		String queryPath = "/test";
		String url="http://localhost:64433/DOESNOTEXIST"+queryPath;
	
		HttpClient hc = gw.getClientFactory().makeHttpClient(new URL(url));
		GetMethod get=new GetMethod(url);
		
		get=new GetMethod(url);
		try{
			hc.executeMethod(get);
			System.out.println(get.getStatusLine());
			int status=get.getStatusCode();
			assertEquals(404, status);
		}
		catch(Exception ex){
			ex.printStackTrace();
			fail();
		}
	}
	
	@FunctionalTest(id="gw_showVersion", description="Tests whether the gateway shows its version on the default page.")
	public void testShowVersion()throws Exception{
		String url="http://localhost:64433/";
		HttpClient hc = gw.getClientFactory().makeHttpClient(new URL(url));
		GetMethod get=new GetMethod(url);
		try{
			hc.executeMethod(get);
			System.out.println(get.getStatusLine());
			int status=get.getStatusCode();
			assertEquals(HttpStatus.SC_OK, status);
			assert(get.getResponseBodyAsString().contains("Version: "+Gateway.VERSION));
		}
		catch(Exception ex){
			ex.printStackTrace();
			fail();
		}
	}
	
	/**
	 * get a minimalistic SOAP message
	 * @param to - WS Addressing To header
	 */
	private byte[] getBody(String to)throws IOException{
		ByteArrayOutputStream bos=new ByteArrayOutputStream();
		OutputStreamWriter wr=new OutputStreamWriter(bos);
		
		wr.write("<soap:Envelope xmlns:soap=\"http://schemas.xmlsoap.org/soap/envelope/\"\n"+
            "xmlns:xsd=\"http://www.w3.org/2001/XMLSchema\" \n"+
            "xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\">"+
           "<soap:Header>");
		wr.write("<wsa:To xmlns:wsa=\"http://www.w3.org/2005/08/addressing\">");
		wr.write(to);
		wr.write("</wsa:To>\n");
		wr.write("</soap:Header>");
		wr.write("<soap:Body>test</soap:Body>");
		wr.write("</soap:Envelope>");
		wr.flush();
		return bos.toByteArray();
	}
	
	/**
	 * get non-SOAP body for POSTing
	 * @param to - WS Addressing To header
	 */
	private byte[] getBodyNoSOAP(String to)throws IOException{
		ByteArrayOutputStream bos=new ByteArrayOutputStream();
		OutputStreamWriter wr=new OutputStreamWriter(bos);
		wr.write("P<soap:Envelope xmlns:soap=\"http://schemas.xmlsoap.org/soap/envelope/\"\n"+
	            "xmlns:xsd=\"http://www.w3.org/2001/XMLSchema\" \n"+
	            "xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\">");
		wr.write("PPP");
		wr.flush();
		return bos.toByteArray();
	}
	
	
	
	private int doRegister(String name, String address)throws Exception{
		String url="http://localhost:64433/VSITE_REGISTRATION_REQUEST";
		HttpClient hc = gw.getClientFactory().makeHttpClient(new URL(url));
		PostMethod post=new PostMethod(url);
		try{
			post.addParameter("name", name);
			post.addParameter("address", address);
			hc.executeMethod(post);
			System.out.println(post.getStatusLine());
			return post.getStatusCode();
		}
		finally{
			post.releaseConnection();
		}
	}

	private boolean containsConsignorIP(List<String>headers){
		for(String s: headers){
			if(s.startsWith(Servlet.X_UNICORE_CONSIGNOR_IP))return true;
		}
		return false;
	}
	
	private static XMLInputFactory xmlInputFactory = XMLInputFactory.newInstance();

	/**
	 * Parses SOAP XML, extracts all headers and inserts them into a set, and returns the SOAP body as a string
	 * @param soap XML to parse
	 * @param headers output set where headers will go
	 * @return SOAP body
	 * @throws Exception
	 */
	private static String parseSoap(String soap, Set<String> headers) throws Exception{
		XMLEventReader reader = xmlInputFactory.createXMLEventReader(new StringReader(soap));
		try{
			SoapVersion soapv = null;
			while (reader.hasNext()) {
				XMLEvent event = reader.nextEvent();
				if (event.isStartElement())
				{
					QName q = event.isStartElement() ? event.asStartElement().getName() :
					event.asEndElement().getName();
					if (Soap11.envelope.equals(q)) {
						soapv = Soap11.getInstance();
					} else if (Soap12.envelope.equals(q)) {
						soapv = Soap12.getInstance();
					} else {
						throw new Exception("(Supported) SOAP Envelope element is not the first element");
					}
					break;
				}
			}
			while(reader.hasNext()){
				XMLEvent event = reader.nextEvent();
				boolean isInHeader = false;
				if (event.isStartElement()){
					QName q = event.isStartElement() ? event.asStartElement().getName() :
					event.asEndElement().getName();
					isInHeader = soapv.getHeader().equals(q);
					int lvl = 0;
					int pointer = reader.peek().getLocation().getCharacterOffset();
					while(reader.hasNext()){
						XMLEvent event2 = reader.nextEvent();
						if (event2.isStartElement()){
							lvl++;
						}
						if (event2.isEndElement()){
							lvl--;
						}
						if(lvl<0) {
							if(!isInHeader){
								int np = event2.getLocation().getCharacterOffset();
								return soap.substring(pointer, np);
							}
							break;
						}
						if(lvl==0 && isInHeader){
							int np = reader.peek().getLocation().getCharacterOffset();
							String header=soap.substring(pointer, np);
							header = header.replaceAll(">\\s+", ">");
							header = header.replaceAll("\\s+<", "<");
							header = header.trim();
							if(!header.isEmpty()) headers.add(header);
							pointer=np;
						}
					}
				}
			}
		} finally {
			reader.close();
		}
		return "";
	}


}
